import { isBrowser, log, MIN_ID } from './misc/util';
import { channel as proto, Message } from './proto';
import { UserMessage } from './service/UserMessage';
export var MAXIMUM_MISSED_HEARTBEAT = 3;
var Channel = (function () {
    function Channel(wc, wsOrDc, type, id, pc) {
        var _this = this;
        log.channel("New Channel " + type + ": Me: " + wc.myId + " with " + id);
        this.wc = wc;
        this.wsOrDc = wsOrDc;
        this.type = type;
        this.pc = pc;
        this.missedHeartbeat = 0;
        this.heartbeatMsg = new Uint8Array(0);
        this.resolveInit = function () { };
        if (isBrowser) {
            wsOrDc.binaryType = 'arraybuffer';
            this.send = this.sendInBrowser;
        }
        else if (!this.pc) {
            this.send = this.sendInNodeOverWebSocket;
        }
        else {
            wsOrDc.binaryType = 'arraybuffer';
            this.send = this.sendInNodeOverDataChannel;
        }
        if (type === Channel.WITH_INTERNAL) {
            this.id = id;
            this.heartbeatMsg = this.createHeartbeatMsg();
            this.init = Promise.resolve();
            this.initHandlers();
        }
        else {
            this.id = MIN_ID;
            if (type === Channel.WITH_JOINING) {
                this.sendInitPing();
            }
            this.init = new Promise(function (resolve, reject) {
                return (_this.resolveInit = function () {
                    _this.heartbeatMsg = _this.createHeartbeatMsg();
                    resolve();
                });
            });
            this.wsOrDc.onmessage = function (_a) {
                var data = _a.data;
                try {
                    var msg = proto.Message.decode(new Uint8Array(data));
                    _this.handleInitMessage(msg);
                }
                catch (err) {
                    log.warn('Decode inner Channel message error: ', err);
                }
            };
        }
    }
    Channel.remoteType = function (type) {
        return type === Channel.WITH_INTERNAL
            ? Channel.WITH_INTERNAL
            : type === Channel.WITH_JOINING
                ? Channel.WITH_MEMBER
                : Channel.WITH_JOINING;
    };
    Object.defineProperty(Channel.prototype, "url", {
        get: function () {
            if (!this.pc) {
                return this.wsOrDc.url;
            }
            return '';
        },
        enumerable: true,
        configurable: true
    });
    Channel.prototype.encodeAndSend = function (_a) {
        var _b = _a === void 0 ? {} : _a, _c = _b.senderId, senderId = _c === void 0 ? this.wc.myId : _c, _d = _b.recipientId, recipientId = _d === void 0 ? 0 : _d, serviceId = _b.serviceId, content = _b.content;
        this.send(Message.encode(Message.create({ senderId: senderId, recipientId: recipientId, serviceId: serviceId, content: content })).finish());
    };
    Channel.prototype.close = function () {
        this.wsOrDc.onmessage = undefined;
        this.wsOrDc.onclose = undefined;
        this.wsOrDc.onerror = undefined;
        if (this.pc) {
            this.pc.oniceconnectionstatechange = function () { };
            this.pc.close();
        }
        else {
            this.wsOrDc.close(1000);
        }
    };
    Channel.prototype.sendHeartbeat = function () {
        this.missedHeartbeat++;
        if (this.missedHeartbeat >= MAXIMUM_MISSED_HEARTBEAT) {
            log.channel(this.id + " channel closed: too many missed heartbeats");
            this.wc.topology.onChannelClose(this);
            this.close();
        }
        else {
            this.send(this.heartbeatMsg);
        }
    };
    Channel.prototype.sendInBrowser = function (data) {
        try {
            this.wsOrDc.send(data);
        }
        catch (err) {
            log.channel('Channel sendInBrowser ERROR', err);
        }
    };
    Channel.prototype.sendInNodeOverWebSocket = function (data) {
        try {
            this.wsOrDc.send(data, { binary: true });
        }
        catch (err) {
            log.channel('Channel sendInNodeOverWebSocket ERROR', err);
        }
    };
    Channel.prototype.sendInNodeOverDataChannel = function (data) {
        this.sendInBrowser(data.slice(0));
    };
    Channel.prototype.handleInitMessage = function (msg) {
        switch (msg.type) {
            case 'initPing': {
                log.channel(this.wc.myId + " received InitPing");
                var _a = msg.initPing, topology = _a.topology, wcId = _a.wcId, senderId = _a.senderId, members = _a.members, key = _a.key;
                if (this.wc.topologyEnum !== topology) {
                    log.channel('Different topology. This message should never be shown: something is wrong');
                }
                else if (members.includes(this.id) && wcId !== this.wc.id) {
                }
                this.wc.id = wcId;
                this.wc.key = key;
                this.id = senderId;
                log.channel(this.wc.myId + " send InitPong");
                this.send(proto.Message.encode(proto.Message.create({ initPong: this.wc.myId })).finish());
                this.initData = { members: members };
                this.initHandlers();
                this.resolveInit();
                break;
            }
            case 'initPong':
                log.channel(this.wc.myId + " received InitPong");
                this.id = msg.initPong;
                this.initHandlers();
                this.resolveInit();
                break;
            default:
                log.channel("Unknown message type: \"" + msg.type + "\". This message should never be shown: something went wrong");
        }
    };
    Channel.prototype.initHandlers = function () {
        var _this = this;
        this.wsOrDc.onmessage = function (_a) {
            var data = _a.data;
            try {
                var msg = Message.decode(new Uint8Array(data));
                if (msg.recipientId === 0 || msg.recipientId === _this.wc.myId) {
                    if (msg.serviceId === UserMessage.SERVICE_ID) {
                        var userData = _this.wc.userMsg.decodeUserMessage(msg.content, msg.senderId);
                        if (userData) {
                            _this.wc.onMessage(msg.senderId, userData);
                        }
                    }
                    else if (msg.serviceId === 0) {
                        _this.missedHeartbeat = 0;
                    }
                    else {
                        _this.wc.streamSubject.next(Object.assign({ channel: _this }, msg));
                    }
                }
                if (msg.recipientId !== _this.wc.myId) {
                    _this.wc.topology.forward(msg);
                }
            }
            catch (err) {
                log.warn('Decode general Channel message error: ', err);
            }
        };
        this.wsOrDc.onclose = function (evt) {
            log.channel("Connection with " + _this.id + " has closed", evt);
            if (_this.pc) {
                _this.pc.oniceconnectionstatechange = function () { };
            }
            _this.wc.topology.onChannelClose(_this);
        };
        if (this.pc) {
            this.pc.oniceconnectionstatechange = function () {
                if (_this.pc && _this.pc.iceConnectionState === 'failed') {
                    _this.close();
                    _this.wc.topology.onChannelClose(_this);
                }
            };
        }
        this.wsOrDc.onerror = function (evt) { return log.channel('Channel error: ', evt); };
    };
    Channel.prototype.createHeartbeatMsg = function () {
        return Message.encode(Message.create({ serviceId: 0, senderId: this.wc.myId, recipientId: this.id })).finish();
    };
    Channel.prototype.sendInitPing = function () {
        log.channel('initialize...');
        this.send(proto.Message.encode(proto.Message.create({
            initPing: {
                topology: this.wc.topologyEnum,
                wcId: this.wc.id,
                senderId: this.wc.myId,
                members: this.wc.members,
                key: this.wc.key,
            },
        })).finish());
    };
    Channel.WITH_INTERNAL = 0;
    Channel.WITH_JOINING = 1;
    Channel.WITH_MEMBER = 2;
    return Channel;
}());
export { Channel };
//# sourceMappingURL=Channel.js.map